/* Copyright (C) 2013-2017 Vincent Forest (vaplv@free.fr)
 *
 * The RSys library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The RSys library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the RSys library. If not, see <http://www.gnu.org/licenses/>. */

/*
 * Generate the cstr_to_list conversion function with respect to the
 * CSTR_LIST_TYPE macro.
 */

#ifndef CSTR_LIST_TYPE
#ifndef CSTR_TO_LIST_H
#define CSTR_TO_LIST_H

#include "cstr.h"
#include "dynamic_array_char.h"

#endif /* CSTR_TO_LIST_H */
#else /* defined(CSTR_LIST_TYPE) */

#ifndef CSTR_LIST_TYPE
  #error "Missing the CSTR_LIST_TYPE macro defining the type of the list"
#endif

#ifndef CSTR_LIST_SUFFIX
  #define CSTR_LIST_SUFFIX CSTR_LIST_TYPE
#endif

res_T
CONCAT(cstr_to_list_, CSTR_LIST_SUFFIX)
  (const char* str,
   const char delimiter,
   CSTR_LIST_TYPE dst[],
   size_t* length,
   const size_t max_length) /* Maximum size of dst */

{
  struct darray_char buf;
  char delim[2] = {'\0', '\0'};
  size_t str_len;
  size_t i;
  char* tk;
  res_T res = RES_OK;

  if(!str) {
    /* Avoid exit block since buf is not initialised */
    return RES_BAD_ARG;
  }
  if(!dst && !length) { /* Useless invocation */
    /* Avoid exit block since buf is not initialised */
    return RES_OK;
  }

  /* Copy str in a temporary buffer to parse */
  darray_char_init(NULL, &buf);
  str_len = strlen(str);
  res = darray_char_resize(&buf, str_len + 1/*null char*/);
  if(res != RES_OK) goto error;
  strncpy(darray_char_data_get(&buf), str, str_len + 1/*null char*/);

  /* Parse the string */
  delim[0] = delimiter;
  tk = strtok(darray_char_data_get(&buf), delim);
  for(i=0; tk; tk = strtok(NULL, delim), ++i) {
    if(dst) {
      if(i >= max_length) {
        res = RES_BAD_ARG;
        goto error;
      }
      res = CONCAT(cstr_to_, CSTR_LIST_SUFFIX)(tk, dst + i);
    } else {
      CSTR_LIST_TYPE d;
      res = CONCAT(cstr_to_, CSTR_LIST_SUFFIX)(tk, &d);
    }
    if(res != RES_OK) goto error;
  }

  if(length)
    *length = i;

exit:
  darray_char_release(&buf);
  return res;
error:
  goto exit;
}

#undef CSTR_LIST_TYPE
#undef CSTR_LIST_SUFFIX

#endif /* defined(CSTR_LIST_TYPE) */

